import cv2
import numpy as np
import math

import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.extractParameters import extractParameters
from zebrazoom.code.preprocessImage import preprocessImage

from ._base import register_tracking_method
from ._fasterMultiprocessingBase import BaseFasterMultiprocessing
from ._tailExtremityTracking import TailTrackingExtremityDetectMixin


class FasterMultiprocessing2(BaseFasterMultiprocessing, TailTrackingExtremityDetectMixin):
  def __init__(self, videoPath, wellPositions, hyperparameters):
    super().__init__(videoPath, wellPositions, hyperparameters)

    # if self._hyperparameters["eyeTracking"]:
      # self._trackingEyesAllAnimalsList = []
    # else:
      # self._trackingEyesAllAnimals = 0
    # for wellNumber in range(0, self._hyperparameters["nbWells"]):
      # if self._hyperparameters["eyeTracking"]:
        # self._trackingHeadingAllAnimalsList.append(np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1, 8)))
      # self._trackingProbabilityOfGoodDetectionList.append(np.zeros((self._hyperparameters["nbAnimalsPerWell"], self._lastFrame-self._firstFrame+1)))

    # if not(self._hyperparameters["nbAnimalsPerWell"] > 1) and not(self._hyperparameters["headEmbeded"]) and (self._hyperparameters["findHeadPositionByUserInput"] == 0) and (self._hyperparameters["takeTheHeadClosestToTheCenter"] == 0):
      # self._trackingProbabilityOfGoodDetectionList = []
    # else:
      # self._trackingProbabilityOfGoodDetectionList = 0
    # self._trackingProbabilityOfGoodDetectionList = []

  def _adjustParameters(self, i, back, frame, initialCurFrame, widgets):
    return None

  def _findTheTwoSides(self, headPosition, bodyContour, curFrame, bestAngle):
    # Finding the 'mouth' of the fish
    unitVector = np.array([math.cos(bestAngle), math.sin(bestAngle)])
    factor     = 1
    headPos    = np.array(headPosition)
    testBorder = headPos + factor * unitVector
    testBorder = testBorder.astype(int)
    while (cv2.pointPolygonTest(bodyContour, (float(testBorder[0]), float(testBorder[1])), True) > 0) and (factor < 100) and (testBorder[0] >= 0) and (testBorder[1] >= 0) and (testBorder[0] < len(curFrame[0])) and (testBorder[1] < len(curFrame)):
      factor = factor + 1
      testBorder = headPos + factor * unitVector

    # Finding the indexes of the two "border points" along the contour (these are the two points that are the closest from the 'mouth' of fish)
    xOtherBorder = testBorder[0]
    yOtherBorder = testBorder[1]
    minDist1 = 1000000000000
    minDist2 = 1000000000000
    indMin1  = 0
    indMin2  = 0
    for i in range(0, len(bodyContour)):
      Pt   = bodyContour[i][0]
      dist = math.sqrt((Pt[0] - xOtherBorder)**2 + (Pt[1] - yOtherBorder)**2)
      if (dist < minDist1):
        minDist2 = minDist1
        indMin2  = indMin1
        minDist1 = dist
        indMin1  = i
      else:
        if (dist < minDist2):
          minDist2 = dist
          indMin2  = i

    return indMin1, indMin2

  def _computeHeading(self, initialContour, lenX, lenY, headPosition):
    xmin = lenX
    ymin = lenY
    xmax = 0
    ymax = 0
    for pt in initialContour:
      if pt[0][0] < xmin:
        xmin = pt[0][0]
      if pt[0][1] < ymin:
        ymin = pt[0][1]
      if pt[0][0] > xmax:
        xmax = pt[0][0]
      if pt[0][1] > ymax:
        ymax = pt[0][1]

    for pt in initialContour:
      pt[0][0] = pt[0][0] - xmin
      pt[0][1] = pt[0][1] - ymin

    headPosition = [headPosition[0] - xmin, headPosition[1] - ymin]

    image = np.zeros((ymax - ymin, xmax - xmin))
    image[:, :] = 255
    image = image.astype(np.uint8)
    kernel  = np.ones((3, 3), np.uint8)
    if type(initialContour) != int:
      cv2.fillPoly(image, pts =[initialContour], color=(0))
      image[:,0] = 255
      image[0,:] = 255
      image[:, len(image[0])-1] = 255
      image[len(image)-1, :]    = 255

    originalShape = 255 - image

    # Heading calculation: first approximation
    minWhitePixel = 1000000000
    bestAngle     = 0
    nTries        = 50
    for i in range(0, nTries):
      angleOption = i * ((2 * math.pi) / nTries)
      startPoint = (int(headPosition[0]), int(headPosition[1]))
      endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))
      testImage  = originalShape.copy()
      testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
      nbWhitePixels = cv2.countNonZero(testImage)

      if nbWhitePixels < minWhitePixel:
        minWhitePixel = nbWhitePixels
        bestAngle     = angleOption
    bestAngleAfterFirstStep = bestAngle

    # Heading calculation: second (and refined) approximation
    # Searching for the optimal value of iterationsForErodeImageForHeadingCalculation
    countTries = 0
    nbIterations2nbWhitePixels = {}
    if "iterationsForErodeImageForHeadingCalculation" in self._hyperparameters:
      iterationsForErodeImageForHeadingCalculation = self._hyperparameters["iterationsForErodeImageForHeadingCalculation"]
    else:
      iterationsForErodeImageForHeadingCalculation = 4
    kernel = np.ones((3, 3), np.uint8)
    nbWhitePixelsMax = 0.3 * cv2.contourArea(initialContour)
    while (iterationsForErodeImageForHeadingCalculation > 0) and (countTries < 50) and not(iterationsForErodeImageForHeadingCalculation in nbIterations2nbWhitePixels):
      testImage2 = cv2.erode(testImage, kernel, iterations = iterationsForErodeImageForHeadingCalculation)
      nbWhitePixels = cv2.countNonZero(testImage2)
      nbIterations2nbWhitePixels[iterationsForErodeImageForHeadingCalculation] = nbWhitePixels
      if nbWhitePixels < nbWhitePixelsMax:
        iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation - 1
      if nbWhitePixels >= nbWhitePixelsMax:
        iterationsForErodeImageForHeadingCalculation = iterationsForErodeImageForHeadingCalculation + 1
      countTries = countTries + 1
    best_iterations = 0
    minDist = 10000000000000
    for iterations in nbIterations2nbWhitePixels:
      nbWhitePixels = nbIterations2nbWhitePixels[iterations]
      dist = abs(nbWhitePixels - nbWhitePixelsMax)
      if dist < minDist:
        minDist = dist
        best_iterations = iterations
    iterationsForErodeImageForHeadingCalculation = best_iterations
    self._hyperparameters["iterationsForErodeImageForHeadingCalculation"] = iterationsForErodeImageForHeadingCalculation

    testImage2 = cv2.erode(originalShape.copy(), kernel, iterations = iterationsForErodeImageForHeadingCalculation)

    maxDist = -1
    for i in range(0, nTries):
      angleOption = bestAngleAfterFirstStep - (math.pi / 5) + i * ((2 * (math.pi / 5)) / nTries)

      startPoint = (int(headPosition[0]), int(headPosition[1]))
      endPoint   = (int(headPosition[0] + 100000 * math.cos(angleOption)), int(headPosition[1] + 100000 * math.sin(angleOption)))

      testImage = testImage2.copy()

      testImage  = cv2.line(testImage, startPoint, endPoint, (0), 1)
      nbWhitePixels = cv2.countNonZero(testImage)
      if nbWhitePixels < minWhitePixel:
        minWhitePixel = nbWhitePixels
        bestAngle     = angleOption

    theta = bestAngle

    if self._hyperparameters["debugHeadingCalculation"]:
      img2 = image.copy()
      img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2RGB)
      cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img2[0])/2 + 20 * math.cos(theta)), int(len(img2)/2 + 20 * math.sin(theta))), (255,0,255), 1)
      self._debugFrame(img2, title='imgForHeadingCalculation')

    return theta + math.pi

# def computeHeading(self, thresh1, x, y):

  # videoWidth  = self._hyperparameters["videoWidth"]
  # videoHeight = self._hyperparameters["videoHeight"]
  # headSize = self._hyperparameters["headSize"]
  # ymin  = y - headSize - 10 if y - headSize >= 0 else 0
  # ymax  = y + headSize + 10 if y + headSize < len(thresh1) else len(thresh1) - 1
  # xmin  = x - headSize - 10 if x - headSize >= 0 else 0
  # xmax  = x + headSize + 10 if x + headSize < len(thresh1[0]) else len(thresh1[0]) - 1
  # img = thresh1[int(ymin):int(ymax), int(xmin):int(xmax)]

  # img[0,:] = 255
  # img[len(img)-1,:] = 255
  # img[:,0] = 255
  # img[:,len(img[0])-1] = 255

  # y2, x2 = np.nonzero(img)
  # x2 = x2 - np.mean(x2)
  # y2 = y2 - np.mean(y2)
  # coords = np.vstack([x2, y2])
  # cov = np.cov(coords)
  # evals, evecs = np.linalg.eig(cov)
  # sort_indices = np.argsort(evals)[::-1]
  # x_v1, y_v1 = evecs[:, sort_indices[0]]  # Eigenvector with largest eigenvalue
  # x_v2, y_v2 = evecs[:, sort_indices[1]]
  # scale = 20
  # theta = self._calculateAngle(0, 0, x_v1, y_v1)
  # theta = (theta - math.pi/2) % (2 * math.pi)

  # if False:
    # width  = len(img[0])
    # height = len(img)
    # option1X = x + width * math.cos(theta)
    # option1Y = y + height * math.sin(theta)
    # option2X = x - width * math.cos(theta)
    # option2Y = y - height * math.sin(theta)
    # if math.sqrt((option1X - width)**2 + (option1Y - height)**2) < math.sqrt((option2X - width)**2 + (option2Y - height)**2):
      # theta += theta + math.pi

  # if self._hyperparameters["debugHeadingCalculation"]:
    # img2 = img.copy()
    # img2 = cv2.cvtColor(img2, cv2.COLOR_GRAY2BGR)
    # cv2.line(img2, (int(len(img2[0])/2), int(len(img2)/2)), (int(len(img[0])/2 + 20 * math.cos(theta)), int(len(img)/2 + 20 * math.sin(theta))), (255,0,0), 1)
    # self._debugFrame(img2, title='imgForHeadingCalculation')

  # return theta

  def _findOptimalIdCorrespondance(self, wellNumber, i):
    from scipy.optimize import linear_sum_assignment

    if i > self._firstFrame:

      costMatrix = np.zeros((len(self._trackingHeadTailAllAnimalsList[wellNumber]), len(self._trackingHeadTailAllAnimalsList[wellNumber])))

      for animalIdPrev in range(0, len(self._trackingHeadTailAllAnimalsList[wellNumber])):
        for animalIdCur in range(0, len(self._trackingHeadTailAllAnimalsList[wellNumber])):
          coordPrevX = self._trackingHeadTailAllAnimalsList[wellNumber][animalIdPrev, i-self._firstFrame-1][0][0]
          coordPrevY = self._trackingHeadTailAllAnimalsList[wellNumber][animalIdPrev, i-self._firstFrame-1][0][1]
          coordCurX  = self._trackingHeadTailAllAnimalsList[wellNumber][animalIdCur,  i-self._firstFrame][0][0]
          coordCurY  = self._trackingHeadTailAllAnimalsList[wellNumber][animalIdCur,  i-self._firstFrame][0][1]
          # TO DO: add some very high cost for (0, 0) coordinates
          costMatrix[animalIdPrev, animalIdCur] = math.sqrt((coordCurX - coordPrevX)**2 + (coordCurY - coordPrevY)**2)

      row_ind, col_ind = linear_sum_assignment(costMatrix)

      return col_ind
    else:
      return np.array([k for k in range(0, len(self._trackingHeadTailAllAnimalsList[wellNumber]))])

  def _switchIdentities(self, correspondance, wellNumber, i):
    trackingHeadTailAllAnimalsListWellNumberOriginal = self._trackingHeadTailAllAnimalsList[wellNumber][:, i-self._firstFrame].copy()
    trackingHeadingAllAnimalsListWellNumberOriginal  = self._trackingHeadingAllAnimalsList[wellNumber][:, i-self._firstFrame].copy()

    for previousId, newId in enumerate(correspondance):
      self._trackingHeadTailAllAnimalsList[wellNumber][previousId, i-self._firstFrame] = trackingHeadTailAllAnimalsListWellNumberOriginal[newId]

    for previousId, newId in enumerate(correspondance):
      self._trackingHeadingAllAnimalsList[wellNumber][previousId, i-self._firstFrame] = trackingHeadingAllAnimalsListWellNumberOriginal[newId]

  def _findCenterByIterativelyDilating(self, initialContour, lenX, lenY):
    x = 0
    y = 0

    xmin = lenX
    ymin = lenY
    xmax = 0
    ymax = 0
    for pt in initialContour:
      if pt[0][0] < xmin:
        xmin = pt[0][0]
      if pt[0][1] < ymin:
        ymin = pt[0][1]
      if pt[0][0] > xmax:
        xmax = pt[0][0]
      if pt[0][1] > ymax:
        ymax = pt[0][1]

    for pt in initialContour:
      pt[0][0] = pt[0][0] - xmin
      pt[0][1] = pt[0][1] - ymin

    image = np.zeros((ymax - ymin, xmax - xmin))
    image[:, :] = 255
    image = image.astype(np.uint8)
    kernel  = np.ones((3, 3), np.uint8)
    if type(initialContour) != int:
      cv2.fillPoly(image, pts =[initialContour], color=(0))
      image[:,0] = 255
      image[0,:] = 255
      image[:, len(image[0])-1] = 255
      image[len(image)-1, :]    = 255
      nbBlackPixels = 1
      dilateIter = 0
      while nbBlackPixels > 0:
        dilateIter   = dilateIter + 1
        dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
        nbBlackPixels = cv2.countNonZero(255-dilatedImage)
      dilateIter   = dilateIter - 1
      dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
      dilatedImage[:,0] = 255
      dilatedImage[0,:] = 255
      dilatedImage[:, len(dilatedImage[0])-1] = 255
      dilatedImage[len(dilatedImage)-1, :]    = 255
      contours, hierarchy = cv2.findContours(dilatedImage, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      maxContour = 0
      maxContourArea = 0
      for contour in contours:
        contourArea = cv2.contourArea(contour)
        if contourArea < int((xmax - xmin) * (ymax - ymin) * 0.8):
          if contourArea > maxContourArea:
            maxContourArea = contourArea
            maxContour     = contour
      M = cv2.moments(maxContour)
      if M['m00']:
        x = int(M['m10']/M['m00'])
        y = int(M['m01']/M['m00'])

    return [x + xmin, y + ymin]

  def _formatOutput(self):
    # if self._hyperparameters["postProcessMultipleTrajectories"]:
      # self._postProcessMultipleTrajectories(self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingProbabilityOfGoodDetectionList[wellNumber])
    if self._auDessusPerAnimalIdList == None:
      return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, 0], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
              for wellNumber in range(self._firstWell, self._lastWell + 1)}
    else:
      return {wellNumber: extractParameters([self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], [], 0, 0, self._auDessusPerAnimalIdList[wellNumber]], wellNumber, self._hyperparameters, self._videoPath, self._wellPositions, self._background)
              for wellNumber in range(self._firstWell, self._lastWell + 1)}

  def run(self):
    self._background = self.getBackground()

    cap = zzVideoReading.VideoCapture(self._videoPath)
    if (cap.isOpened()== False):
      print("Error opening video stream or file")
    
    lastFrameRememberedForBackgroundExtract = 0
    
    # if self._hyperparameters["backgroundSubtractorKNN"]:
      # fgbg = cv2.createBackgroundSubtractorKNN()
      # for i in range(0, min(self._lastFrame - 1, 500), int(min(self._lastFrame - 1, 500) / 10)):
        # cap.set(1, min(self._lastFrame - 1, 500) - i)
        # ret, frame = cap.read()
        # fgmask = fgbg.apply(frame)
      # cap.release()
      # cap = zzVideoReading.VideoCapture(videoPath)

    i = self._firstFrame

    if self._firstFrame:
      cap.set(1, self._firstFrame)

    previousFrames = None
    widgets = None
    while (i < self._lastFrame + 1):

      if (self._hyperparameters["freqAlgoPosFollow"] != 0) and (i % self._hyperparameters["freqAlgoPosFollow"] == 0):
        print("Tracking: frame:",i)
        if self._hyperparameters["popUpAlgoFollow"]:
          from zebrazoom.code.popUpAlgoFollow import prepend

          prepend("Tracking: frame:" + str(i))

      if self._hyperparameters["debugTracking"]:
        print("frame:",i)

      ret, frame = cap.read()

      if ret:
        
        if self._hyperparameters["invertBlackWhiteOnImages"]:
          frame = 255 - frame
        
        if self._hyperparameters["imagePreProcessMethod"]:
          frame = preprocessImage(frame, self._hyperparameters)
        
        # if self._hyperparameters["backgroundSubtractorKNN"]:
          # frame = fgbg.apply(frame)
          # frame = 255 - frame

        for wellNumber in range(self._firstWell, self._lastWell + 1):

          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
          xtop = self._wellPositions[wellNumber]['topLeftX']
          ytop = self._wellPositions[wellNumber]['topLeftY']
          lenX = self._wellPositions[wellNumber]['lengthX']
          lenY = self._wellPositions[wellNumber]['lengthY']
          # if self._hyperparameters["backgroundSubtractorKNN"]:
            # grey = frame
          # else:
          grey = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

          curFrame = grey[ytop:ytop+lenY, xtop:xtop+lenX].copy()
          initialCurFrame = curFrame.copy()
          # if not(self._hyperparameters["backgroundSubtractorKNN"]):
          back = self._background[ytop:ytop+lenY, xtop:xtop+lenX]
          putToWhite = ( curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
          curFrame[putToWhite] = 255
          # else:
            # self._hyperparameters["paramGaussianBlur"] = int(math.sqrt(cv2.countNonZero(255 - curFrame) / self._hyperparameters["nbAnimalsPerWell"]) / 2) * 2 + 1
          # if self._hyperparameters["paramGaussianBlur"]:
            # blur = cv2.GaussianBlur(curFrame, (self._hyperparameters["paramGaussianBlur"], self._hyperparameters["paramGaussianBlur"]),0)
          # else:
            # blur = curFrame
          headPositionFirstFrame = 0

          ret, thresh1 = cv2.threshold(curFrame.copy(), 254, 255, cv2.THRESH_BINARY)
          
          thresh1[:,0] = 255
          thresh1[0,:] = 255
          thresh1[:, len(thresh1[0])-1] = 255
          thresh1[len(thresh1)-1, :]    = 255
          
          contours, hierarchy = cv2.findContours(thresh1,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
          areas = np.array([cv2.contourArea(contour) for contour in contours])

          maxIndexes = []
          for numFish in range(0, self._hyperparameters["nbAnimalsPerWell"]):
            maxArea = -1
            maxInd  = -1
            for idx, area in enumerate(areas):
              if area > maxArea and area > 0.7 * self._hyperparameters["minAreaBody"] and area < 1.3 * self._hyperparameters["maxAreaBody"]:
                maxArea = area
                maxInd  = idx
            areas[maxInd] = -1
            if maxInd != -1:
              maxIndexes.append(maxInd)

          for animal_Id, idx in enumerate(maxIndexes):
            bodyContour = contours[idx]
            M = cv2.moments(bodyContour)
            if M['m00']:
              # x = int(M['m10']/M['m00'])
              # y = int(M['m01']/M['m00'])
              # headPosition = [x, y]
              headPosition = self._findCenterByIterativelyDilating(bodyContour.copy(), len(curFrame[0]), len(curFrame))

              self._trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame][0][0] = headPosition[0]
              self._trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame][0][1] = headPosition[1]

              heading = self._computeHeading(bodyContour.copy(), len(curFrame[0]), len(curFrame), headPosition)

              self._trackingHeadingAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame] = heading

              if self._hyperparameters["trackTail"] == 1 :

                res = self._findTheTwoSides(headPosition, bodyContour, curFrame, heading)

                # Finding tail extremity
                rotatedContour = bodyContour.copy()
                rotatedContour = self._rotate(rotatedContour,int(headPosition[0]),int(headPosition[1]),heading)
                debugAdv = False

                [MostCurvyIndex, distance2] = self._findTailExtremete(rotatedContour, bodyContour, headPosition[0], int(res[0]), int(res[1]), debugAdv, curFrame, self._hyperparameters["tailExtremityMaxJugeDecreaseCoeff"])

                # Getting Midline
                if self._hyperparameters["detectMouthInsteadOfHeadTwoSides"] == 0:
                  tail = self._getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, curFrame, self._nbTailPoints-1, distance2, debugAdv)
                else:
                  tail = self._getMidline(int(res[0]), int(res[1]), int(MostCurvyIndex), bodyContour, curFrame, self._nbTailPoints, distance2, debugAdv)
                  tail = np.array([tail[0][1:len(tail[0])]])
                tail = np.insert(tail, 0, headPosition, axis=1)
                self._trackingHeadTailAllAnimalsList[wellNumber][animal_Id, i-self._firstFrame] = tail

          # Eye tracking for frame i
          # if self._hyperparameters["eyeTracking"]:
            # self._eyeTracking(animalId, i, frame, thresh1, self._trackingHeadingAllAnimalsList[wellNumber], self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber])

          correspondance = self._findOptimalIdCorrespondance(wellNumber,  i)

          self._switchIdentities(correspondance, wellNumber, i)

          self._debugTracking(i, self._trackingHeadTailAllAnimalsList[wellNumber], self._trackingHeadingAllAnimalsList[wellNumber], curFrame)

          if self._hyperparameters["updateBackgroundAtInterval"]:
            self._updateBackgroundAtInterval(i, wellNumber, initialCurFrame, self._trackingHeadTailAllAnimalsList[wellNumber], initialCurFrame)
          
          if ("updateBackgroundAtIntervalRememberLastFrame" in self._hyperparameters) and (self._hyperparameters["updateBackgroundAtIntervalRememberLastFrame"]):
            lastFrameRememberedForBackgroundExtract = self._updateBackgroundAtIntervalRememberLastFrame(i, wellNumber, grey, lastFrameRememberedForBackgroundExtract)
            
          
          if self._hyperparameters["freqAlgoPosFollow"]:
            if i % self._hyperparameters["freqAlgoPosFollow"] == 0:
              print("Tracking at frame", i)

        if self._hyperparameters["detectMovementWithRawVideoInsideTracking"]:
          previousFrames = self._detectMovementWithRawVideoInsideTracking(i, grey, previousFrames)

      paramsAdjusted = self._adjustParameters(i, back, frame, initialCurFrame, widgets)
      if paramsAdjusted is not None:
        i, widgets = paramsAdjusted
        cap.set(1, i)
      else:
        i = i + 1

    cap.release()
    return self._formatOutput()


register_tracking_method('fasterMultiprocessing2', FasterMultiprocessing2)



import numpy as np
import pandas as pd

from ._openResultsFile import openResultsFile


def createExcelFileWithRawData(videoName: str, numWell: int, numAnimal: int, startFrame: int, endFrame: int) -> dict:
  with openResultsFile(videoName, 'r+') as results:
        
    firstFrame = results.attrs['firstFrame']
    lastFrame = results.attrs['lastFrame']
    if startFrame < firstFrame or endFrame > lastFrame:
      raise ValueError(f"[{startFrame}, {endFrame}] is not a valid interval, tracking was run from frame {results.attrs['firstFrame']} to frame {results.attrs['lastFrame']}")
    animalPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}'
    if animalPath not in results:
      raise ValueError(f"data for animal {numAnimal} in well {numWell} doesn't exist")
    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results, cannot calculate kinematic parameters')
    if 'videoPixelSize' not in results.attrs:
      raise ValueError(f'videoPixelSize not found in the results, cannot calculate kinematic parameters')

    dataGroup = results[animalPath]['dataPerFrame']
    boutData = {'AnimalNumber': numAnimal, 'BoutStart': startFrame, 'BoutEnd': endFrame}
    start = startFrame - firstFrame
    end = endFrame + 1 - firstFrame
    HeadPos = dataGroup['HeadPos'][start:end]
    TailPosX = dataGroup['TailPosX'][start:end]
    TailPosY = dataGroup['TailPosY'][start:end]
    boutData['HeadX'] = HeadPos['X']
    boutData['HeadY'] = HeadPos['Y']
    boutData['TailX_VideoReferential'] = np.column_stack([HeadPos['X']] + [TailPosX[col] for col in dataGroup['TailPosX'].attrs['columns']])
    boutData['TailY_VideoReferential'] = np.column_stack([HeadPos['Y']] + [TailPosY[col] for col in dataGroup['TailPosY'].attrs['columns']])
    boutData['TailAngle_Raw'] = dataGroup['TailAngle'][start:end]
    boutData['Heading'] = dataGroup['Heading'][start:end]
    
    ### Saving raw movement data in csv file
    if 'TailAngle_smoothed' in dataGroup:
      if not(len(dataGroup["TailAngle_smoothed"].shape)) or not(type(dataGroup["TailAngle_smoothed"][0]) == np.float):
        del dataGroup["TailAngle_smoothed"]
        dataGroup.create_dataset("TailAngle_smoothed", (len(dataGroup['HeadPos']),), dtype=float)
    else:
      dataGroup.create_dataset("TailAngle_smoothed", (len(dataGroup['HeadPos']),), dtype=float)
    dataGroup["TailAngle_smoothed"][:] = 0
    for numBout in range(results[animalPath]['listOfBouts'].attrs["numberOfBouts"]):
      BoutStart = results[animalPath]['listOfBouts']['bout'+str(numBout)].attrs["BoutStart"]
      BoutEnd   = results[animalPath]['listOfBouts']['bout'+str(numBout)].attrs["BoutEnd"]
      if ((BoutStart <= startFrame) and (startFrame <= BoutEnd)) or ((BoutStart <= endFrame) and (endFrame <= BoutEnd)):
        dataGroup['TailAngle_smoothed'][BoutStart-1:BoutEnd] = results[animalPath]['listOfBouts']['bout'+str(numBout)]['TailAngle_smoothed'][:]
    # Getting tail length
    tailLength = []
    for i in range(len(TailPosX)):
      tailLength.append(np.sum(np.sqrt(np.diff([val for val in TailPosX[i]])**2 + np.diff([val for val in TailPosY[i]])**2)))
    tailLength = np.array(tailLength)
    # Getting subsequent points distance
    subsequentPointsDistance = np.array([np.sqrt(np.sum((np.array([val for val in HeadPos[i+1]]) - np.array([val for val in HeadPos[i]]))**2)).tolist() for i in range(len(HeadPos)-1)] + [0])
    
    # Export to csv
    movementDataToExport= pd.DataFrame(np.transpose(np.array([dataGroup['TailAngle'][start:end], dataGroup['TailAngle_smoothed'][start:end], dataGroup['Heading'][start:end], tailLength, subsequentPointsDistance])), columns=['TailAngle', 'TailAngle_smoothed', 'Heading', 'TailLength', 'subsequentPointsDistance'])
    movementDataToExport.to_csv('animal' + str(numAnimal) + '_frame_' + str(startFrame) + '_to_' + str(endFrame) + '.csv')
    
    return movementDataToExport



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.tailTrackTry4InitialPerpendicularDirections import tailTrackTry4InitialPerpendicularDirections
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.dualDirectionTailDetection import dualDirectionTailDetection
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.tailTrackFindNextPoint import tailTrackFindNextPoint
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import calculateAngle
from zebrazoom.code.extractParameters import calculateTailAngle
from scipy.interpolate import interp1d
from numpy import linspace
import numpy as np
import math

def trackTail(self, frameROI, headPosition, hyperparameters, wellNumber, frameNumber, lastFirstTheta):
  
  steps   = hyperparameters["steps"]
  nbList  = 10 if hyperparameters["nbList"] == -1 else hyperparameters["nbList"]
  maxDepth = hyperparameters["maxDepth"]

  angle = 0

  points = np.zeros((2, 0))
  
  debug = hyperparameters["debugHeadEmbededFindNextPoints"]
  if debug:
    print("Frame number:", frameNumber + self._firstFrame)
  
  lenX = len(frameROI[0]) - 1
  lenY = len(frameROI) - 1
  
  if "tries4rotationsCombination" in hyperparameters and hyperparameters["tries4rotationsCombination"]:
    (points, lastFirstTheta, medianPixTotList) = tailTrackTry4InitialPerpendicularDirections(headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY)
  elif "dualDirectionTailDetection" in hyperparameters and hyperparameters["dualDirectionTailDetection"]:
    points = np.insert(points, 0, headPosition, 1)
    (points, lastFirstTheta, medianPixTotList) = dualDirectionTailDetection(headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY)
    if len(points) == 0: # or medianPixTotList > hyperparameters["maximumMedianValueOfAllPointsAlongTheTail"]:
      return [], lastFirstTheta
  else:
    (points, lastFirstTheta, medianPixTotList) = tailTrackFindNextPoint(0, headPosition, frameROI, points, lastFirstTheta, maxDepth, steps, nbList,  hyperparameters, debug, lenX, lenY)
  
  points = np.insert(points, 0, headPosition, axis=1)
  
  if False:
    points = np.transpose(points)
    num_samples = hyperparameters["nbTailPoints"]
    x = points[:, 0]
    y = points[:, 1]
    path_length = np.sum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
    desired_distance = path_length / (num_samples - 1)
    cumulative_distances = np.cumsum(np.sqrt(np.diff(x)**2 + np.diff(y)**2))
    cumulative_distances = np.insert(cumulative_distances, 0, 0)
    interp_func = interp1d(cumulative_distances, points, axis=0)
    new_distances = np.linspace(0, path_length, num=num_samples)
    resampled_points = interp_func(new_distances)
    points = np.transpose(resampled_points)
  
  output = np.zeros((1, hyperparameters["nbTailPoints"], 2))
  for idx, x in enumerate(points[0]):
    if idx < hyperparameters["nbTailPoints"]:
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]
  
  if len(points[0]) < hyperparameters["nbTailPoints"]:
    for idx in range(len(points[0]), hyperparameters["nbTailPoints"]):
      output[0][idx][0] = output[0][len(points[0])-1][0]
      output[0][idx][1] = output[0][len(points[0])-1][1]
  else:
    if len(points[0]) > hyperparameters["nbTailPoints"]:
      output[0][hyperparameters["nbTailPoints"]-1][0] = points[0][len(points[0])-1]
      output[0][hyperparameters["nbTailPoints"]-1][1] = points[1][len(points[0])-1]
  
  return output, lastFirstTheta



import math
import os

import numpy as np
import pandas as pd

from ._openResultsFile import openResultsFile


def createDistanceBetweenFramesExcelFile(videoName: str) -> None:
  with openResultsFile(videoName, 'r') as results:
    numberOfWells = len(results['wellPositions'])
    firstFrame = results.attrs['firstFrame']
    lastFrame = results.attrs['lastFrame']

    if 'videoFPS' not in results.attrs:
      raise ValueError(f'videoFPS not found in the results')
    if 'videoPixelSize' not in results.attrs:
      raise ValueError(f'videoPixelSize not found in the results')
    fps = results.attrs['videoFPS']
    pixelSize = results.attrs['videoPixelSize']

    data = pd.DataFrame(columns=['wellId', 'frameNumber', 'second', 'minute', 'xPosition', 'yPosition', 'instantaneousDistance'], index=range((lastFrame - firstFrame) * numberOfWells))
    data.loc[:, 'frameNumber'] = [x for _ in range(numberOfWells) for x in range(firstFrame, lastFrame)]
    data.loc[:, 'second'] = data['frameNumber'] / fps
    data.loc[:, 'minute'] = data['second'] / 60
    data.loc[:, 'instantaneousDistance'] = 0
    for wellIdx in range(numberOfWells):
      data.loc[data.index[wellIdx * (lastFrame - firstFrame): (wellIdx + 1) * (lastFrame - firstFrame)], 'wellId'] = wellIdx
      for bout in results[f'dataForWell{wellIdx}']['dataForAnimal0']['listOfBouts']:
        boutStart = results[f'dataForWell{wellIdx}']['dataForAnimal0']['listOfBouts'][bout].attrs['BoutStart'] - firstFrame
        boutEnd = results[f'dataForWell{wellIdx}']['dataForAnimal0']['listOfBouts'][bout].attrs['BoutEnd'] - firstFrame + 1
        rowSlice = np.s_[wellIdx * (lastFrame - firstFrame) + boutStart: wellIdx * (lastFrame - firstFrame) + boutEnd]
        data.loc[data.index[rowSlice], 'xPosition'] = results[f'dataForWell{wellIdx}']['dataForAnimal0']['dataPerFrame']['HeadPos']['X'][boutStart:boutEnd]
        data.loc[data.index[rowSlice], 'yPosition'] = results[f'dataForWell{wellIdx}']['dataForAnimal0']['dataPerFrame']['HeadPos']['Y'][boutStart:boutEnd]
    ZZoutputFolder = os.path.dirname(results.filename)

  # Calculating the "instantaneousDistance"
  previousNumWell = 0
  previousXPos = -1
  previousYPos = -1
  for i in range(0, len(data) - 1):
    print(i, len(data))
    if data.loc[i, 'wellId'] == previousNumWell:
      if not(np.isnan(data.loc[i, 'xPosition'])):
        if not(previousXPos == -1) and not(previousYPos == -1):
          data.loc[i, 'instantaneousDistance'] = math.sqrt((data.loc[i, 'xPosition'] - previousXPos)**2 + (data.loc[i, 'yPosition'] - previousYPos)**2) * pixelSize
        previousXPos = data.loc[i, 'xPosition']
        previousYPos = data.loc[i, 'yPosition']
    else:
      previousNumWell = data.loc[i, 'wellId']
      previousXPos = -1
      previousYPos = -1

  # Changing the shape of the data
  finalDataArray = 0
  for wellId in range(0, numberOfWells):
    wellData = data.loc[data['wellId'] == wellId]
    if type(finalDataArray) == int:
      finalDataArray = np.concatenate((np.transpose(np.array([[i for i in range(firstFrame, lastFrame)]])), np.transpose(np.array([[i/fps for i in range(firstFrame, lastFrame)]])), np.transpose(np.array([wellData['instantaneousDistance']])), np.transpose(np.array([wellData['xPosition']])), np.transpose(np.array([wellData['yPosition']]))), axis=1)
    else:
      finalDataArray = np.concatenate((finalDataArray, np.transpose(np.array([wellData['instantaneousDistance']])), np.transpose(np.array([wellData['xPosition']])), np.transpose(np.array([wellData['yPosition']]))), axis=1)

  # Converting to excel
  finalData = pd.DataFrame(data=finalDataArray, columns=['frame', 'second'] + [item for i in range(0, numberOfWells) for item in ['distanceWell' + str(i), 'xPosition' + str(i), 'yPosition' + str(i)]])
  finalData.to_excel(os.path.join(ZZoutputFolder, videoName + '_Distances.xlsx'))



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.trackTail import trackTail
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.getNewFrameROI import getNewFrameROI
from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.removeAnimalJustTrackedFromFrame import removeAnimalJustTrackedFromFrame
import numpy as np
import cv2

def animalDetectionBackToWholeImage(self, k, frame, wellNumber, animalNotTracked):
  
  listOfAnimalIds = []
  for i in range(len(animalNotTracked)):
    if animalNotTracked[i]:
      if self._trackingDataPerWell[wellNumber][i][k-1][0][0] == 0 and self._trackingDataPerWell[wellNumber][i][k-1][0][1] == 0:
        listOfAnimalIds = [i] + listOfAnimalIds
      else:
        listOfAnimalIds.append(i)
  
  curIdInListOfAnimalIds = 0
  for i in range(len(animalNotTracked)):
    if animalNotTracked[i]:
      
      # Animal Id selection
      animalId = listOfAnimalIds[curIdInListOfAnimalIds]
      curIdInListOfAnimalIds += 1
      
      # Frame ROI selection
      wellXtop = self._wellPositions[wellNumber]['topLeftX']
      wellYtop = self._wellPositions[wellNumber]['topLeftY']
      lenghtWell_X = self._wellPositions[wellNumber]['lengthX']
      lenghtWell_Y = self._wellPositions[wellNumber]['lengthY']
      frameROI = frame[wellYtop:wellYtop+lenghtWell_Y, wellXtop:wellXtop+lenghtWell_X]
      
      # Applying gaussian filter
      paramGaussianBlur = self._hyperparameters["paramGaussianBlur"]
      frameROI = cv2.GaussianBlur(frameROI, (paramGaussianBlur, paramGaussianBlur), 0)
      
      # Tracking
      (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(frameROI)
      if not(minVal >= self._hyperparameters["minimumHeadPixelValue"]):
        a, self._lastFirstTheta[wellNumber] = trackTail(self, frameROI, headPosition, self._hyperparameters, wellNumber, k, self._lastFirstTheta[wellNumber])
        if len(a):
          a[0, :, 0] -= self._wellPositions[wellNumber]['topLeftX']
          a[0, :, 1] -= self._wellPositions[wellNumber]['topLeftY']

          self._trackingDataPerWell[wellNumber][animalId][k][:len(a[0])] = a
          
          frameROI = getNewFrameROI(self, k, frame, wellNumber, animalId)
          
          if np.sum(np.sum(255 - frameROI)) < self._hyperparameters["minimumPixelIntensitySumForAnimalDetection"]:
            
            self._trackingDataPerWell[wellNumber][animalId][k] = self._trackingDataPerWell[wellNumber][animalId][k-1]
          
          else:
          
            # 'Removing' animal just tracked
            if self._hyperparameters["nbAnimalsPerWell"] > 1:
              frame = removeAnimalJustTrackedFromFrame(self, frame, wellNumber, animalId, k)